package com.hurlant.jsobject
{
	import flash.external.ExternalInterface;
	import flash.utils.ByteArray;
	import flash.utils.Proxy;
	import flash.utils.flash_proxy;

	use namespace flash_proxy;

	/**
	 * 
	 * A proxy for the browser DOM, to let
	 * ecmascript code in the player feel just like it's 
	 * running with the browser environment.
	 * 
	 * @author henri
	 * 
	 */
	dynamic public class JSObject extends Proxy
	{
		internal var state:Object;
		internal var items:Array;
		private static var lock:Lock;
		 
		
		
		public function JSObject(obj:Object, l:Lock) {
			if (l==null) {
				throw "JSObject has a private constructor.";
			}
			state=obj;
		}
	
		static private var window:JSObject = null;	
		static private function init():void {
			lock = new Lock;
			ExternalInterface.addCallback("jsobject", jsCall);
			window = new JSObject(ExternalInterface.call("HURLANT_jsobject_dispatch", "setup", ExternalInterface.objectID), lock);
		}
		static private var asfuncs:Object = {};
		static private var asobjs:Object = {};
		static private function jsCall(n:String, that:*, a:*):* {
//			trace("JSObject: jsCall("+n+", "+that+", "+a+")");
			var f:Function = asfuncs[n];
			if (f!=null) {
				for (var i:int=0;i<a.length;i++) {
					var ref:Object = a[i];
					a[i] = resolve(ref); 
				}
				var ret:* = f.apply(resolve(that), a);
				return wrap(ret);
			}
			return null;
		}
		static public function getWindow():JSObject {
			if (lock == null) {
				init();
			}
			return window;
		}
		static private var objMap:Object = {};
		static private function getFromRef(ref:Object):JSObject {
			if (objMap[ref.id]==null) {
//				trace("JSObject: wrap in JSObject id "+ref.id);
				objMap[ref.id] = new JSObject(ref, lock);
			}
			return objMap[ref.id];
		}
		
		static private var idcount:int=0;
		static private function genId():String {
			return "aso"+(idcount++)
		}
		
		static private function getRef(name:String, parent:JSObject = null):* {
			var obj:Object = ExternalInterface.call("HURLANT_jsobject_dispatch", "getReference", name, parent?parent.state.id:null);
			return resolve(obj);
		}
		static private function setRef(name:String, parent:JSObject, value:*):void {
			ExternalInterface.call("HURLANT_jsobject_dispatch", "setRef", name, parent.state.id, wrap(value));
		}
		static private function wrap(value:*):* {
			if (value is JSObject) {
//				trace("JSObject: wrap("+value+")=>unwrap to "+value.state.id);
				return {type:"ref", id:value.state.id};
			} else if (value is Function) {
				// check if it's a wrapped jsobject
				try { // this can fail, and we're still okay.
					if (value.jsobject!=null) {
						return wrap(value.jsobject);
					}
				} catch (e:Error) {}
				var id:String = genId();
				asfuncs[id] = value;
//				trace("JSObject: wrap("+value+")="+id);
				return {type:"func", id:id};
			} else {
				if (value is String) {
					value = value.replace(/\\/g, "\\\\");
				}
				var t:String=typeof value;
				if ((value!=null)&&(!(value is Array))&&(typeof value=="object")) {
					id = genId();
					asobjs[id] = value;
					var str:String;
					if (value is ByteArray) {
						str ="[object ByteArray]"; // ByteArray::toString() outputs things ExternalInterface can't cope with. :(
					} else {
						str = value.toString();
					}
					return {type:"obj", id:id, str:str};
				}
				// XXX we should look at objects here and map to AS ids.
				return {type:"value", val:value};
			}
		}
		static private function callRef(name:String, parent:JSObject, args:Array):* {
			for (var i:int=0;i<args.length;i++) {
				args[i] = wrap(args[i]);
			}
			var v:* = ExternalInterface.call("HURLANT_jsobject_dispatch", "callRefById", name, parent.state.id, args);
			return resolve(v);
		}
		static private function delRef(name:String, parent:JSObject):Boolean {
			return ExternalInterface.call("HURLANT_jsobject_dispatch", "delReference", name, parent.state.id);
		}
		static private function hasRef(name:String, parent:JSObject):Boolean {
			return ExternalInterface.call("HURLANT_jsobject_dispatch", "hasReference", name, parent.state.id);
		}
		static private function getIndexes(parent:JSObject):Array {
			return ExternalInterface.call("HURLANT_jsobject_dispatch", "getIndexes", parent.state.id);
		}
		static private function wrapJSFunc(obj:JSObject):* {
			var f:Object = function wrapper(...args):* {
				for (var i:int=0;i<args.length;i++) {
					args[i] = wrap(args[i]);
				}
				if (this.constructor == wrapper) {
					return resolve(ExternalInterface.call("HURLANT_jsobject_dispatch", "newRef", obj.state.id, args));
				} else {
					return resolve(ExternalInterface.call("HURLANT_jsobject_dispatch", "callRef", obj.state.id, args));
				}
			};
			f.toString = function():String {
				return obj.toString();
			};
			f.jsobject = obj;
			f.setPropertyIsEnumerable("toString", false);
			f.setPropertyIsEnumerable("jsobject", false);
			return f;
		}
		static private function wrapASFunc(v:Function):* {
			var f:Object = function(a:*):* {
				for (var i:int=0;i<a.length;i++) {
					var ref:Object = a[i];
					a[i] = resolve(ref); 
				}
				var ret:* = v.apply(this,a);
				return wrap(ret);
			}
			return f;
		}
		static private function resolve(ref:Object):* {
			if (ref===null) return null;
			var t:String = typeof ref.value;
			if (t != "undefined") {
				if (t == "string") {
					// strip the "X" ... "X" from the string.
					return ref.value.slice(1, -1);
				} else 
					return ref.value;
			}
			var o:JSObject = getFromRef(ref);
			if (ref.callable == true) {
				return wrapJSFunc(o);
			}
			return o;
		}
		
		// the public API
		
		override flash_proxy function getProperty(name:*):* {
//			trace("JSObject: "+state.id+".getProperty("+name+")");
			return getRef(name, this);
		}
		
		override flash_proxy function setProperty(name:*, value:*):void {
//			trace("JSObject: "+state.id+".setRef("+name+", "+value+")");
			setRef(name, this, value);
		}
		
		override flash_proxy function callProperty(name:*, ...rest):* {
//			trace("JSObject: "+state.id+".callRef("+name+", "+rest+")");
			return callRef(name, this, rest);
		}
		
		override flash_proxy function deleteProperty(name:*):Boolean {
//			trace("JSObject: "+state.id+".delRef("+name+")");
			return delRef(name, this);
		}
		
		override flash_proxy function hasProperty(name:*):Boolean {
//			trace("JSObject: "+state.id+".hasRef("+name+")");
			return hasRef(name, this);
		}
		
		override flash_proxy function nextNameIndex(index:int):int {
//			trace("JSObject: "+state.id+".nextNameIndex("+index+")");
			// initial call
			if (index==0) {
				items = getIndexes(this);
			}
			if (index < items.length) {
				return index+1;
			} else {
				return 0;
			}
		}
		override flash_proxy function nextName(index:int):String {
//			trace("JSObject: "+state.id+".nextName("+index+")");
			return items[index-1];
		}
		override flash_proxy function nextValue(index:int):* {
//			trace("JSObject: "+state.id+".nextValue("+index+")");
			return getProperty(nextName(index));
		}
		
	}
}
class Lock {}